/* jpeg.c : pour dcoder un fichier JPEG (JPG->BMP) */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * UltimateTetris                           Copyright (C) 2001, AnTiKoNs *
 *----------------/                                                      *
 * This program is free software; you can redistribute it and/or modify  *
 * it under the terms of the GNU General Public License as published by  *
 * the Free Software Foundation; either version 2 of the License, or     *
 * (at your option) any later version.                                   *
 *                                                                       *
 * This program is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 * GNU General Public License for more details.                          *
 *                                                                       *
 * The author of this program may be contacted at antikons@ifrance.com   *
 *                                                                       *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// ---------------------------------------------------------------------------
//
// This software is based in part on the work of the Independent JPEG Group
// IJG libjpeg : copyright (C) 1991-1996, Thomas G. Lane
//
// ---------------------------------------------------------------------------

#define	STRICT
#define	WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <stdio.h>
#include <malloc.h>
#include <setjmp.h>

#include "jpeg_lib.h"
#include "jpeg_error.h"

#include "../debug.h"

// ---------------------------------------------------------------------------

//******************* JPEG DECOMPRESSION SAMPLE INTERFACE ********************
//
// This half of the example shows how to read data from the JPEG decompressor.
// It's a bit more refined than the above, in that we show:
//   (a) how to modify the JPEG library's standard error-reporting behavior;
//   (b) how to allocate workspace using the library's memory manager.
//
// Just to make this example a little different from the first one, we'll
// assume that we do not intend to put the whole image into an in-memory
// buffer, but to send it line-by-line someplace else.  We need a one-
// scanline-high JSAMPLE array as a work buffer, and we will let the JPEG
// memory manager allocate it for us.  This approach is actually quite useful
// because we don't need to remember to deallocate the buffer separately: it
// will go away automatically when the JPEG object is cleaned up.

// ERROR HANDLING:
//
// The JPEG library's standard error handler (jerror.c) is divided into
// several "methods" which you can override individually.  This lets you
// adjust the behavior without duplicating a lot of code, which you might
// have to update with each future release.
//
// Our example here shows how to override the "error_exit" method so that
// control is returned to the library's caller when a fatal error occurs,
// rather than calling exit() as the standard error_exit method does.
//
// We use C's setjmp/longjmp facility to return control.  This means that the
// routine which calls the JPEG library must first execute a setjmp() call to
// establish the return point.  We want the replacement error_exit to do a
// longjmp().  But we need to make the setjmp buffer accessible to the
// error_exit routine.  To do this, we make a private extension of the
// standard JPEG error handler object.  (If we were using C++, we'd say we
// were making a subclass of the regular error handler.)

// Here's the extended error handler struct:
struct	my_error_mgr
{
	struct	jpeg_error_mgr	pub;			// "public" fields
	jmp_buf					setjmp_buffer;	// for return to caller
};

typedef	struct	my_error_mgr	*my_error_ptr;

// Here's the routine that will replace the standard error_exit method:
METHODDEF(void)	my_error_exit(j_common_ptr	cinfo)
{
	// cinfo->err really points to a my_error_mgr struct, so coerce pointer
	my_error_ptr myerr=(my_error_ptr) cinfo->err;

	// Always display the message.
	// We could postpone this until after returning, if we chose.
	(*cinfo->err->output_message)(cinfo);

	// Return control to the setjmp point
	longjmp(myerr->setjmp_buffer,1);
}

// Dcompression JPEG
// Retourne : TRUE en cas de succs, FALSE en cas d'erreur
BOOL JpegLoadFromFile	(
						char	*filename,	// Nom du fichier
						HGLOBAL	*phmemBuf,	// Buffer de l'image
						int		*piWidth,	// Largeur de l'image
						int		*piHeight,	// Hauteur de l'image
						int		*piBpp,		// Nombre de bits/pixel de l'image
						DWORD	*pdwSize	// Taille du buffer
						)
{
	// This struct contains the JPEG decompression parameters and pointers to
	// working space (which is allocated as needed by the JPEG library).
	struct	jpeg_decompress_struct	cinfo;
	// We use our private extension JPEG error handler.
	// Note that this struct must live as long as the main JPEG parameter
	// struct, to avoid dangling-pointer problems.
	struct	my_error_mgr		jerr;
	FILE						*infile;			// source file
	JSAMPARRAY					buffer;				// Output row buffer
	int							row_stride,			// phys row w/in output buff
								iAlignedRowStride;	// Alignement BMP
	LPSTR						pBuf;				// Buffer

	// In this example we want to open the input file before doing anything else
	// so that the setjmp() error recovery below can assume the file is open.
	// VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
	// requires it in order to read binary files.
	if((infile=fopen(filename,"rb"))==NULL) 
	{
		return	FALSE;
	}

	// Step 1: allocate and initialize JPEG decompression object

	// We set up the normal JPEG error routines, then override error_exit.
	cinfo.err=jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit=my_error_exit;
	// Establish the setjmp return context for my_error_exit to use.
	if(setjmp(jerr.setjmp_buffer))
	{
		Error("DDLoadJpeg : LoadJPGImage : JpegLoadFromFile : Cannot read \"%s\" (Jpeg internal error)",filename);
		// If we get here, the JPEG code has signaled an error.
		// We need to clean up the JPEG object, close the input file, and return
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		return	FALSE;
	}

	// Now we can initialize the JPEG decompression object.
	jpeg_create_decompress(&cinfo);

	// Step 2: specify data source (eg, a file)
	jpeg_stdio_src(&cinfo,infile);

	// Step 3: read file parameters with jpeg_read_header()
	(void)jpeg_read_header(&cinfo,TRUE);
	// We can ignore the return value from jpeg_read_header since
	//   (a) suspension is not possible with the stdio data source, and
	//   (b) we passed TRUE to reject a tables-only JPEG file as an error.
	// See libjpeg.doc for more info.

	// Step 4: set parameters for decompression

	// In this example, we don't need to change any of the defaults set by
	// jpeg_read_header(), so we do nothing here.

	// Step 5: Start decompressor
	(void)jpeg_start_decompress(&cinfo);
	// We can ignore the return value since suspension is not possible
	// with the stdio data source.

	// We may need to do some setup of our own at this point before reading
	// the data.  After jpeg_start_decompress() we have the correct scaled
	// output image dimensions available, as well as the output colormap
	// if we asked for color quantization.
	// In this example, we need to make an output work buffer of the right size.

	// JSAMPLEs per row in output buffer
	row_stride=cinfo.output_width*cinfo.output_components;
	// Make a one-row-high sample array that will go away when done with image 
	buffer=(*cinfo.mem->alloc_sarray)((j_common_ptr)&cinfo,JPOOL_IMAGE,
				row_stride,1);

	*piWidth=cinfo.output_width;
	*piHeight=cinfo.output_height;
	*piBpp=8*cinfo.output_components;
	// On alloue assez pour l'alignement DWORD. Le surplus n'a pas  tre
	// rempli heureusement ! La formule magique provient de Micro$oft...
	iAlignedRowStride=(((*piWidth*(*piBpp)+31)&~31)>>3);
	*pdwSize=*piHeight*iAlignedRowStride;
	*phmemBuf=GlobalAlloc(GMEM_MOVEABLE,*pdwSize);

	if(!*phmemBuf)
	{
		Error("DDLoadJpeg : LoadJPGImage : JpegLoadFromFile : File \"%s\" is too large; not enough memory (%lu bytes needed)",filename,*pdwSize);
		
		// Mnage
		(void)jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
		fclose(infile);
		return	FALSE;
	}

	pBuf=GlobalLock(*phmemBuf);

	// Step 6: while (scan lines remain to be read)
	//           jpeg_read_scanlines(...);

	// Here we use the library's state variable cinfo.output_scanline as the
	// loop counter, so that we don't have to keep track ourselves.
	while(cinfo.output_scanline<cinfo.output_height)
	{
		// jpeg_read_scanlines expects an array of pointers to scanlines.
		// Here the array is only one element long, but you could ask for
		// more than one scanline at a time if that's more convenient.
		(void)jpeg_read_scanlines(&cinfo,buffer,1);
		// Assume put_scanline_someplace wants a pointer and sample count.

		// On formatte  la Windows (cd BGR, il parait que a remonte  PM !)
		// Le fichier JPEG retourne du RGB (comme tout le monde) et Windows
		// attend du BGR...
		if(*piBpp>8)
		{
			char	*p=(char*)buffer[0],
					r,b;
			int		l;

			for(l=0;l<row_stride;l+=3)
			{
				r=*p;
				b=*(p+2);

				*p=b;
				*(p+2)=r;

				p+=3;
			}
		}

		// On recopie dans le buffer
		CopyMemory(pBuf,buffer[0],row_stride);
		// On saute  la ligne suivante (avec alignement !)
		pBuf+=iAlignedRowStride;
	}

	GlobalUnlock(*phmemBuf);

	// Step 7: Finish decompression
	(void)jpeg_finish_decompress(&cinfo);
	// We can ignore the return value since suspension is not possible
	// with the stdio data source.

	// Step 8: Release JPEG decompression object

	// This is an important step since it will release a good deal of memory.
	jpeg_destroy_decompress(&cinfo);

	// After finish_decompress, we can close the input file.
	// Here we postpone it until after no more JPEG errors are possible,
	// so as to simplify the setjmp error logic above.  (Actually, I don't
	// think that jpeg_destroy can do an error exit, but why assume anything...)
	fclose(infile);

	// At this point you may want to check to see whether any corrupt-data
	// warnings occurred (test whether jerr.pub.num_warnings is nonzero).

	// And we're done!
	return	TRUE;
}

// SOME FINE POINTS:
//
// In the above code, we ignored the return value of jpeg_read_scanlines,
// which is the number of scanlines actually read.  We could get away with
// this because we asked for only one line at a time and we weren't using
// a suspending data source.  See libjpeg.doc for more info.
//
// We cheated a bit by calling alloc_sarray() after jpeg_start_decompress();
// we should have done it beforehand to ensure that the space would be
// counted against the JPEG max_memory setting.  In some systems the above
// code would risk an out-of-memory error.  However, in general we don't
// know the output image dimensions before jpeg_start_decompress(), unless we
// call jpeg_calc_output_dimensions().  See libjpeg.doc for more about this.
//
// Scanlines are returned in the same order as they appear in the JPEG file,
// which is standardly top-to-bottom.  If you must emit data bottom-to-top,
// you can use one of the virtual arrays provided by the JPEG memory manager
// to invert the data.  See wrbmp.c for an example.
//
// As with compression, some operating modes may require temporary files.
// On some systems you may need to set up a signal handler to ensure that
// temporary files are deleted if the program is interrupted.  See libjpeg.doc.

struct user_data
{
	char *begin;
	char *cur;
	char *end;
};

struct jpeg_decompress_struct_and_user_data
{
	struct jpeg_decompress_struct jds;
	struct user_data ud;
};


METHODDEF(void) buffer_init_source(j_decompress_ptr cinfo)
{
}

METHODDEF(boolean) buffer_fill_input_buffer(j_decompress_ptr cinfo)
{
	char *cur,*end;
	struct jpeg_decompress_struct_and_user_data	*pjdsaud = (struct jpeg_decompress_struct_and_user_data	*)cinfo;
	cur = pjdsaud->ud.cur;
	end = pjdsaud->ud.end;
	cinfo->src->next_input_byte = cur;
	cinfo->src->bytes_in_buffer = end - cur;
	return 1;
}

METHODDEF(void) buffer_skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
	struct jpeg_decompress_struct_and_user_data	*pjdsaud = (struct jpeg_decompress_struct_and_user_data	*)cinfo;
	pjdsaud->ud.cur += num_bytes;
}

METHODDEF(boolean) buffer_resync_to_restart(j_decompress_ptr cinfo, int desired)
{
	return jpeg_resync_to_restart(cinfo,desired);
}

METHODDEF(void) buffer_term_source(j_decompress_ptr cinfo)
{
}

BOOL JpegLoadFromBuffer	(
						char	*pbuffer,	// pointer to buffer
						int buffer_size,
						HGLOBAL	*phmemBuf,	// Buffer de l'image
						int		*piWidth,	// Largeur de l'image
						int		*piHeight,	// Hauteur de l'image
						int		*piBpp,		// Nombre de bits/pixel de l'image
						DWORD	*pdwSize	// Taille du buffer
						)
{
	// This struct contains the JPEG decompression parameters and pointers to
	// working space (which is allocated as needed by the JPEG library).
	j_decompress_ptr	cinfo;
	// We use our private extension JPEG error handler.
	// Note that this struct must live as long as the main JPEG parameter
	// struct, to avoid dangling-pointer problems.
	struct	my_error_mgr		jerr;
	JSAMPARRAY					buffer;				// Output row buffer
	int							row_stride,			// phys row w/in output buff
								iAlignedRowStride;	// Alignement BMP
	LPSTR						pBuf;				// Buffer
	struct jpeg_source_mgr		srcmgr;
	struct jpeg_decompress_struct_and_user_data	jdsaud;

	srcmgr.next_input_byte = NULL;
	srcmgr.bytes_in_buffer = 0;
	srcmgr.init_source = buffer_init_source;
	srcmgr.fill_input_buffer = buffer_fill_input_buffer;
	srcmgr.skip_input_data = buffer_skip_input_data;
	srcmgr.resync_to_restart = buffer_resync_to_restart;
	srcmgr.term_source = buffer_term_source;
	
	jdsaud.ud.begin = jdsaud.ud.cur = pbuffer;
	jdsaud.ud.end = pbuffer + buffer_size;
	
	// Step 1: allocate and initialize JPEG decompression object
	cinfo = (j_decompress_ptr)&jdsaud;

	// We set up the normal JPEG error routines, then override error_exit.
	cinfo->err=jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit=my_error_exit;
	// Establish the setjmp return context for my_error_exit to use.
	if(setjmp(jerr.setjmp_buffer))
	{
		Error("DDLoadJpeg : LoadJPGImage : JpegLoadFromBuffer : Jpeg internal error");
		// If we get here, the JPEG code has signaled an error.
		// We need to clean up the JPEG object, close the input file, and return
		jpeg_destroy_decompress(cinfo);
		free(cinfo);
		return	FALSE;
	}

	// Now we can initialize the JPEG decompression object.
	jpeg_create_decompress(cinfo);

	// Step 2: specify data source (eg, a file)
	cinfo->src = &srcmgr;
	
	// Step 3: read file parameters with jpeg_read_header()
	(void)jpeg_read_header(cinfo,TRUE);
	// We can ignore the return value from jpeg_read_header since
	//   (a) suspension is not possible with the stdio data source, and
	//   (b) we passed TRUE to reject a tables-only JPEG file as an error.
	// See libjpeg.doc for more info.

	// Step 4: set parameters for decompression

	// In this example, we don't need to change any of the defaults set by
	// jpeg_read_header(), so we do nothing here.

	// Step 5: Start decompressor
	(void)jpeg_start_decompress(cinfo);
	// We can ignore the return value since suspension is not possible
	// with the stdio data source.

	// We may need to do some setup of our own at this point before reading
	// the data.  After jpeg_start_decompress() we have the correct scaled
	// output image dimensions available, as well as the output colormap
	// if we asked for color quantization.
	// In this example, we need to make an output work buffer of the right size.

	// JSAMPLEs per row in output buffer
	row_stride=cinfo->output_width*cinfo->output_components;
	// Make a one-row-high sample array that will go away when done with image 
	buffer=(*cinfo->mem->alloc_sarray)((j_common_ptr)cinfo,JPOOL_IMAGE,
				row_stride,1);

	*piWidth=cinfo->output_width;
	*piHeight=cinfo->output_height;
	*piBpp=8*cinfo->output_components;
	// On alloue assez pour l'alignement DWORD. Le surplus n'a pas  tre
	// rempli heureusement ! La formule magique provient de Micro$oft...
	iAlignedRowStride=(((*piWidth*(*piBpp)+31)&~31)>>3);
	*pdwSize=*piHeight*iAlignedRowStride;
	*phmemBuf=GlobalAlloc(GMEM_MOVEABLE,*pdwSize);

	if(!*phmemBuf)
	{
		Error("DDLoadJpeg : LoadJPGImage : JpegLoadFromBuffer : File is too large; not enough memory (%lu bytes needed)",*pdwSize);
		
		// Mnage
		(void)jpeg_finish_decompress(cinfo);
		jpeg_destroy_decompress(cinfo);
		free(cinfo);
		return	FALSE;
	}

	pBuf=GlobalLock(*phmemBuf);

	// Step 6: while (scan lines remain to be read)
	//           jpeg_read_scanlines(...);

	// Here we use the library's state variable cinfo.output_scanline as the
	// loop counter, so that we don't have to keep track ourselves.
	while(cinfo->output_scanline<cinfo->output_height)
	{
		// jpeg_read_scanlines expects an array of pointers to scanlines.
		// Here the array is only one element long, but you could ask for
		// more than one scanline at a time if that's more convenient.
		(void)jpeg_read_scanlines(cinfo,buffer,1);
		// Assume put_scanline_someplace wants a pointer and sample count.

		// On formatte  la Windows (cd BGR, il parait que a remonte  PM !)
		// Le fichier JPEG retourne du RGB (comme tout le monde) et Windows
		// attend du BGR...
		if(*piBpp>8)
		{
			char	*p=(char*)buffer[0],
					r,b;
			int		l;

			for(l=0;l<row_stride;l+=3)
			{
				r=*p;
				b=*(p+2);

				*p=b;
				*(p+2)=r;

				p+=3;
			}
		}

		// On recopie dans le buffer
		CopyMemory(pBuf,buffer[0],row_stride);
		// On saute  la ligne suivante (avec alignement !)
		pBuf+=iAlignedRowStride;
	}

	GlobalUnlock(*phmemBuf);

	// Step 7: Finish decompression
	(void)jpeg_finish_decompress(cinfo);
	// We can ignore the return value since suspension is not possible
	// with the stdio data source.

	// Step 8: Release JPEG decompression object

	// This is an important step since it will release a good deal of memory.
	jpeg_destroy_decompress(cinfo);
	
	// At this point you may want to check to see whether any corrupt-data
	// warnings occurred (test whether jerr.pub.num_warnings is nonzero).
	
	// And we're done!
	return	TRUE;
}
